/*
** Module   : ConsoleHelper.java
** Abstract : The class provides JNI calls to text console based C code. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130723 Created initial version.
*/

package com.goldencode.p2j.ui.client.chui.driver.console;

/**
 * The ConsoleHelper class provides the interface to the "ncurses" library functions or other.
 * implementation of the character based terminal functions.
 * <p>
 * The default colors are white foreground and black background or other currently used by OS
 * console terminal re-definitions.
 */
public class ConsoleHelper 
{
   private static ConsoleHelper instance;
   
   /**
    * This flag is true if the system property "console.color" has been set.
    */
   public static final boolean isColorEnabled =
      (System.getProperty("console.color") != null);

   /**
    * Default empty constructor.
    */
   private ConsoleHelper()
   {
   }

   /**
    * Initialize the native layer.
    */
   public static void initChui()
   {
      System.loadLibrary("p2j");
      initConsole(); // call native function to initialize console.
   }

   /**
    * Access the instance of the console helper.
    *
    * @return   The static console helper instance.
    */
   public synchronized static ConsoleHelper getDefaultConsoleHelper()
   {
      if (instance == null)
      {
         instance = new ConsoleHelper();
      }

      return instance;
   }
   
   /**
    * This method is used for getting the line-drawing special characters from underlying native
    * layer.
    *
    * @param index 
    *        The index of the character to return.
    *
    * @return   The character at the given index from internal array.
    */
   public static native int getACSchar(int index);

   /**
    * This method is used for getting the attributes from underlying native layer.
    *
    * @param index 
    *        The index of the attribute to return.
    *
    * @return   The attribute at the given index from internal array.
    */
   public static native int getAttribute(int index);

   /**
    * This method performs console initialization on the native side.
    */
   private static native void initConsole();

   /**
    * Perform a bulk output operation for the whole screen at once with
    * output masking when desired. Set the cursor afterwards.
    * <p>
    * The specified array safe for use with characters that are UNICODE
    * encoded which are also outside of the 7-bit ASCII range. Such
    * characters are passed through to the native layer directly and are
    * not converted to a natively encoded form.  This could be provided
    * but since one of the primary uses of this method is for drawing
    * characters (e.g. <code>ACS_HLINE</code>) which are encoded using the
    * full 32-bits, any attempt to convert these into a single byte native
    * encoding would lose data.
    * <p>
    * Because of the above limitation, DO NOT USE THIS FOR OUTPUT OF USER
    * DATA THAT COULD CONTAIN DATA OUTSIDE OF THE 7-bit ASCII RANGE.
    *
    * @param    chr
    *           The array of characters and attributes to output. The array
    *           is supposed to provide an item per every screen position.
    *           This item is either -1 if this position is masked or a code
    *           which goes directly to ncurses - a combination of character,
    *           attribute and color.
    * @param    cursorX
    *           final cursor position, column
    * @param    cursorY
    *           final cursor position, row
    */
   public native void addArrayNative(int chr[], int cursorX, int cursorY);

   /**
    * Add a string to the virtual terminal at the current cursor position.
    *
    * @param    str
    *           The array of characters to output.
    * @param    len
    *           The lenght of the string to draw in characters.
    * @param    attrib
    *           The attribute to be used while drawing.
    * @param    colorpair
    *           The color value to be used while drawing.
    */
   public native void addStringNative(byte[] str, int len, int attrib, int colorpair);

   /**
    * Beeps the tone on the terminal.
    */
   public native void beep();

   /**
    * Clears the current console screen.
    */
   public native void clear();

   /**
    * Gets the number of columns on the screen.
    *
    * @return   The number of columns (in characters).
    */
   public native int getScreenColumnsNative();

   /**
    * Gets the number of rows on the screen.
    *
    * @return   The number of rows (lines of characters).
    */
   public native int getScreenRowsNative();
   
   /**
    * Check if the terminal is able to display colors or not.
    *
    * @return   <code>true</code> if display is color capable, <code>false</code> otherwise.
    */
   public native boolean hasColors();

   /**
    * Set absolute cursor position
    * 
    * @param    x
    *           New X coordinate of the cursor.
    * @param    y
    *           New Y coordinate of the cursor.
    *
    * @return   <code>true</code> if cursor positioning succeeded.
    */
   public native boolean setCursorNative(int x, int y);

   /**
    * Synchronizes the state of the console buffers with native terminal window. Meaning the kind
    * of refresh calls depending on the backend used.
    */
   public native void syncNative();
   
   /**
    * Sets the terminal type used by console backend.
    *
    * @param    name
    *           The new terminal type (must not be <code>null</code>).
    */
   public static native void setTermType(String name);
   
   /**
    * Gets the terminal type reported by console native backend.
    *
    * @return   The terminal type or an empty string on any failure.
    */
   public static native String getTermType();
   
   /**
    * Set the cursor visibility on or off. Note that can invisible cursor
    * can still be moved (with {@link #setCursorNative} and used normally.  This
    * method only affects the visibility of the cursor.
    *
    * @param    on
    *           When <code>true</code> the cursor will become visible, when
    *           <code>false</code> the cursor will become invisible.
    *
    * @return   The previous state of the cursor (<code>true</code> means visible).
    */
   public native boolean setCursorStatus(boolean on)
   throws IllegalStateException;

   /**
    * This method reads the key from keyboard by the special reading thread.
    *
    * @return   The integer code of the key has been read or -1 if there is no character to read.
    */
   public native int readKey();

   /**
    * Get X coordinate of the cursor.
    *
    * @return   The integer value of the X coordinate.
    */
   public native int getx();

   /**
    * Get Y coordinate of the cursor.
    *
    * @return   The integer value of the Y coordinate.
    */
   public native int gety();

   /**
    * Re-activates the console mode after ending of the previous suspension state.
    */
   public native void resume();

   /**
    * Save the current state of the console mode and switch to TTY mode temporarily.
    */
   public native void suspend();

   /**
    * Reset terminal settings.  May be called only if suspend() was previously called. 
    */
   public native void resetMode();
}
