/*
** Module   : KeyProcessor.java
** Abstract : Java standard header
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SVG 20050609   @21446 Initial version.
** 002 SIY 20050831   @22541 Formatting and spelling fixes.
** 003 SIY 20060209   @24459 Cleaned up imports.
** 004 NVS 20060213   @24586 Fixed bug in Esc key processing.
** 005 SIY 20060817   @28647 Fixed another bug in Esc key processing.
**                           The source key code should be remapped before
**                           applying ESC flag.
** 006 SIY 20080530   @38460 Added workaround for VT220 DELETE key.
**                           Instead of sending usual 127 code it sends
**                           long sequence which is handled incorrectly
**                           because keyboard mapper does not expect to see
**                           long ESC-sequences and does not know that
**                           this is just DELETE key.
**                           Ref: #38459
** 007 SIY 20100619          Switching to own implementation of character UI.
** 008 SIY 20100804          Changes in imports, some refactorings.
** 009 SIY 20110112          Changes related to low level driver refactoring.
** 010 SIY 20110927          Updated javadoc.
** 011 GES 20111005          Moved location to a more appropriate package.
** 012 EVL 20130725          Moved to location of the new CHUI console driver.
*/

package com.goldencode.p2j.ui.client.chui.driver.console;


import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;

/**
 * Process raw keys from driver and convert them into P2J compatible codes.
 * <p>
 * The keys generated by the driver are expected in the NCURSES-like format,
 * but to get Progress 4GL compatible key codes additional transformation
 * is required.<br> 
 * Following transformations are performed: <br>
 * 
 * -The ESC key handled separately. When it appears in the input 
 * from the driver, is is assumed that new ESC sequence is started. When 
 * this happens further key codes (up to 4 in some special cases) are
 * collected in order to form complete sequence. When sequence is collected
 * it is translated into one of the Progress 4GL keys for further processing.
 * <br>
 * 
 * - The ESC-&lt;ASCII letter&gt; sequences are translated into 
 * Ctrl+&lt;ASCII letter&gt; code by converting letter to upper case and adding 
 * 1024 (which marks pressed Control key). 
 * Note that translation slightly varies from Progress here, because both, 
 * lower and upper case letters are handled identically as upper case letters.<br>
 * 
 * - Translated sequence then is checked for presence in the {@link Keyboard} 
 * class internal mappings and if sequence is not present there it is rejected
 * and -1 is returned. If check is passed then resulting key is returned as result
 * of mapping.
 * <br>
 * - Special handling for the ESC-(91, 51, 126) sequence is performed on the 
 * VT220 terminal. This sequence is handled as DELETE key if workaround is 
 * enabled.<br>
 * - Special keys (see OLD_VK_* constants below) are mapped into 
 * appropriate Progress 4GL/P2J keys ({@link Key}.VK_* constants)<br>
 * - ENTER and CR are translated as VK_ENTER key.<br>
 * Refer to {@link Keyboard} class for more details about relationship between
 * key names and Progress 4GL/P2J codes.
 */
public class KeyProcessor
{
   // -----------------------------------------------
   // Original codes
   // -----------------------------------------------
   public static final int  OLD_VK_TAB        = '\t';
   public static final int  OLD_VK_ESCAPE     = 0x1b;
   public static final int  OLD_VK_CTRL_G     = 0x07;
   public static final int  OLD_VK_CTRL_U     = 0x15;
   public static final int  OLD_VK_CTRL_Z     = 0x1a;
   public static final int  OLD_VK_DOWN       = 0402;
   public static final int  OLD_VK_UP         = 0403;
   public static final int  OLD_VK_LEFT       = 0404;
   public static final int  OLD_VK_RIGHT      = 0405;
   public static final int  OLD_VK_HOME       = 0406;
   public static final int  OLD_VK_BACK_SPACE = 0407;
   public static final int  OLD_VK_F1         = 0411;
   public static final int  OLD_VK_F2         = 0412;
   public static final int  OLD_VK_F3         = 0413;
   public static final int  OLD_VK_F4         = 0414;
   public static final int  OLD_VK_F5         = 0415;
   public static final int  OLD_VK_F6         = 0416;
   public static final int  OLD_VK_F7         = 0417;
   public static final int  OLD_VK_F8         = 0420;
   public static final int  OLD_VK_F9         = 0421;
   public static final int  OLD_VK_F10        = 0422;
   public static final int  OLD_VK_F11        = 0423;
   public static final int  OLD_VK_F12        = 0424;
   public static final int  OLD_VK_F13        = 0425;
   public static final int  OLD_VK_F14        = 0426;
   public static final int  OLD_VK_F15        = 0427;
   public static final int  OLD_VK_F16        = 0430;
   public static final int  OLD_VK_F17        = 0431;
   public static final int  OLD_VK_F18        = 0432;
   public static final int  OLD_VK_F19        = 0433;
   public static final int  OLD_VK_F20        = 0434;
   public static final int  OLD_VK_DELETE     = 0512;
   public static final int  OLD_VK_INSERT     = 0513;
   public static final int  OLD_VK_PAGE_DOWN  = 0522;
   public static final int  OLD_VK_PAGE_UP    = 0523;
   public static final int  OLD_VK_ENTER      = 0527;
   public static final int  OLD_VK_BACK_TAB   = 0541;
   public static final int  OLD_VK_END        = 0550;

   /** Flag which enables/disables VT220 DELETE key handling.*/
   private boolean handleVT220Delete = false;

   /** VT220 sends this sequence (after leading ESC) for DEL key. */
   private int[] sequence = new int[] {91, 51, 126};

   /** Pointer in the sequence array. */
   private int seqPtr = 0;

   /**
    * Was ESC key previously pressed. This may mean direct ESC usage or ALT
    * pressing with some other key.
    */
   private boolean escPressed = false;

   /**
    * Constructor with the empty body. This class doesn't need a complex init
    * for an instance.
    */
   public KeyProcessor()
   {
   }

   /**
    * Maps curses key code to VK_ constants from <code>ProgressKeys</code>
    * class. This method is called only from keyboard reading thread so thread
    * safe is not required.
    *
    * @param   key
    *          Original key code from curses.
    *
    * @return  Key code constant or (-1) if this key should be ignored.
    */
   public int map(int key)
   {
      int jk = -1;

      if (escPressed)
      {
         // special handling of DEL sequence on VT200 terminal
         if (handleVT220Delete && sequence[seqPtr] == key)
         {
            seqPtr++;

            if (seqPtr == sequence.length)
            {
               seqPtr = 0;
               escPressed = false;
               return Key.VK_DELETE;
            }

            return -1;
         }

         // this is the key following the ESC
         escPressed = false;

         // Remapping lowercase Latin to uppercase. This differs
         // with Progress but seems to be more convenient to user because
         // Ctrl+<letter> combinations are case insensitive (and even
         // keyboard layout insensitive). Note the usage of direct
         // calculation not Character class methods. This is
         // because we are dealing with integer key codes not characters so
         // this is dirty hack anyway
         if (key >= 'a' && key <= 'z')
         {
            key -= 'a' - 'A';
         }

         switch(key)
         {
            case Key.VK_ESCAPE:
               jk = -1;
               break;

            default:
               jk = translate(key) | 1024;
         }

         if (jk != -1 && Keyboard.keyLabel(jk).length() == 0)
            jk = -1;
      }
      else
      {
         // this is a regular mode key
         if (key == Key.VK_ESCAPE)
         {
            escPressed = true;
         }
         else
         {
            jk = translate(key);
         }
      }

      // reset automaton
      seqPtr = 0;

      return jk;
   }

   /**
    * Enable/disable VT220 DELETE key workaround.
    *
    * @param   enable
    *          <code>true</code> - enable workaround.
    */
   public void setDelWorkaroundActive(boolean enable)
   {
      handleVT220Delete = enable;
   }

   /**
    * Key translation worker.
    *
    * @param   key
    *          Raw keystroke from NCURSES
    *
    * @return  key translated into P2J key codes.
    */
   private int translate(int key)
   {
      switch (key)
      {
         case '\n' :
         case '\r' :
            key = Key.VK_ENTER;
            break;
         case '\b' :
            key = Key.VK_BACKSPACE;
            break;

         case OLD_VK_BACK_SPACE :
            key = Key.VK_BACKSPACE;
            break;

         case OLD_VK_ENTER:
            key = Key.VK_ENTER;
            break;

         case OLD_VK_TAB:
            key = Key.VK_TAB;
            break;

         case OLD_VK_DOWN:
            key = Key.VK_CURSOR_DOWN;
            break;

         case OLD_VK_UP:
            key = Key.VK_CURSOR_UP;
            break;

         case OLD_VK_LEFT:
            key = Key.VK_CURSOR_LEFT;
            break;

         case OLD_VK_RIGHT:
            key = Key.VK_CURSOR_RIGHT;
            break;

         case OLD_VK_HOME:
            key = Key.VK_HOME;
            break;

         case OLD_VK_F1:
            key = Key.VK_F1;
            break;

         case OLD_VK_F2:
            key = Key.VK_F2;
            break;

         case OLD_VK_F3:
            key = Key.VK_F3;
            break;

         case OLD_VK_F4:
            key = Key.VK_F4;
            break;

         case OLD_VK_F5:
            key = Key.VK_F5;
            break;

         case OLD_VK_F6:
            key = Key.VK_F6;
            break;

         case OLD_VK_F7:
            key = Key.VK_F7;
            break;

         case OLD_VK_F8:
            key = Key.VK_F8;
            break;

         case OLD_VK_F9:
            key = Key.VK_F9;
            break;

         case OLD_VK_F10:
            key = Key.VK_F10;
            break;

         case OLD_VK_F11:
            key = Key.VK_F11;
            break;

         case OLD_VK_F12:
            key = Key.VK_F12;
            break;

         case OLD_VK_F13:
            key = Key.VK_F13;
            break;

         case OLD_VK_F14:
            key = Key.VK_F14;
            break;

         case OLD_VK_F15:
            key = Key.VK_F15;
            break;

         case OLD_VK_F16:
            key = Key.VK_F16;
            break;

         case OLD_VK_F17:
            key = Key.VK_F17;
            break;

         case OLD_VK_F18:
            key = Key.VK_F18;
            break;

         case OLD_VK_F19:
            key = Key.VK_F19;
            break;

         case OLD_VK_F20:
            key = Key.VK_F20;
            break;

         case OLD_VK_DELETE:
            key = Key.VK_DELETE;
            break;

         case OLD_VK_INSERT:
            key = Key.VK_INSERT;
            break;

         case OLD_VK_PAGE_DOWN:
            key = Key.VK_PAGE_DOWN;
            break;

         case OLD_VK_PAGE_UP:
            key = Key.VK_PAGE_UP;
            break;

         case OLD_VK_BACK_TAB:
            key = Key.VK_BACK_TAB;
            break;

         case OLD_VK_END:
            key = Key.VK_END;
            break;
      }
      return key;
   }
}
