/*
** Module   : terminal.c
** Abstract : Generic implementation for character mode-based console functions.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130723 Created initial version to implement text console terminal JNI.
*/

#include "terminal.h"

#include "com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper.h"

// These globals will be initialized in initConsole() function
int base_rows = 0;
int base_cols = 0;

/** 
 * This function outputs the screenful of data at once using an array of codes where every
 * character position is either a skip indicator (-1) or a fully prepared code with character,
 * attribute and color included. The final step is the cursor positioning.
 *
 * @param   env      pointer to a structure that provides access to JNI environment.
 *
 * @param   jo       reference to the Java object who called this function.
 *
 * @param   arr      the array of characters and attributes to output. The array is supposed to
 *                   provide an item per every screen position.  This item is either -1 if this
 *                   position is masked or a code which goes directly to ncurses - a combination
 *                   of character, attribute and color.
 *
 * @param   cx       final cursor position, column.
 *
 * @param   cy       final cursor position, row.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_addArrayNative
  (JNIEnv *env, jobject jo, jintArray arr, jint cx, jint cy)
{
   addArrayNative(env, arr, cx, cy);
}

/**
 * Add a string to the virtual terminal at the current cursor position.
 *
 * @param   env        pointer to a structure that provides access to JNI environment.
 *
 * @param   jo         reference to the Java object who called this function.
 *
 * @param   jstr       The array of characters to output.
 * 
 * @param   len        the lenght of the string to draw in characters.
 *
 * @param   attrib     the attribute to be used while drawing.
 *
 * @param   colorpair  the color value to be used while drawing.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_addStringNative
  (JNIEnv *env, jobject jo, jbyteArray jstr, jint len, jint attrib, jint colorpair)
{
   addStringNative(env, jstr, len, attrib, colorpair);
}

/**
 * Beeps the tone on the terminal.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_beep
  (JNIEnv *env, jobject jo)
{
   beepNative();
}

/**
 * Clears the current console screen.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_clear
  (JNIEnv *env, jobject jo)
{
   clearNative();
}

/**
 * This method is used for getting the line-drawing special characters from underlying native
 * layer.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   class   reference to the Java class who called this function (from static context).
 *
 * @param   index   The index of the character to return. 
 *
 * @return   The character at the given index from internal array.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getACSchar
  (JNIEnv *env, jclass class, jint index)
{
   return getACSchar(index);
}

/**
 * This method is used for getting the attributes from underlying native layer.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   class   reference to the Java class who called this function (from static context).
 *
 * @param   index   the index of the attribute to return. 
 *
 * @return   The attribute at the given index from internal array.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getAttribute
  (JNIEnv *env, jclass class, jint index)
{
   return getAttribute(index);
}

/**
 * Return the number of columns that the screen can display.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The number of columns (in characters).
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getScreenColumnsNative
  (JNIEnv *env, jobject jo)
{
   return (jint) base_cols;
}

/**
 * Return the number of rows that the screen can display.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The number of rows (lines of characters).
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getScreenRowsNative
  (JNIEnv *env, jobject jo)
{
   return (jint) base_rows;
}

/**
 * Gets the terminal type reported by console native backend.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   cls     reference to the Java class who called this function (from static context).
 *
 * @return   The terminal type or an empty string on any failure.
 */
JNIEXPORT jstring JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getTermType
  (JNIEnv *env, jclass  cls)
{
   return getTermType(env);
}

/**
 * Get X coordinate of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The integer value of the X coordinate.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_getx
  (JNIEnv *env, jobject jo)
{
   return getx();
}

/**
 * Get Y coordinate of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The integer value of the Y coordinate.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_gety
  (JNIEnv *env, jobject jo)
{
   return gety();
}

/**
 * Check if the terminal is able to display colors or not.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   <code>true</code> if display is color capable, <code>false</code> otherwise.
 */
JNIEXPORT jboolean JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_hasColors
  (JNIEnv *env, jobject jo)
{
   return hasColors();
}

/**
 * This method performs console initialization on the native side.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_initConsole
  (JNIEnv *env, jobject jo)
{
   initConsole(env);
}

/**
 * This method reads the key from keyboard by the special reading thread.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The integer code of the key that was read or -1 if there is no character to read.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_readKey
  (JNIEnv *env, jobject jo)
{
   return readKey();
}

/**
 * Reset terminal settings.  May be called only if suspend() was previously called. 
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_resetMode
  (JNIEnv *env, jobject jo)
{
   resetMode();
}

/**
 * Re-activates the console mode after ending of the previous suspension state.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_resume
  (JNIEnv *env, jobject jo)
{
   resume();
}

/**
 * Set absolute cursor position
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 * 
 * @param   x       new X coordinate of the cursor.
 *
 * @param   y       new Y coordinate of the cursor.
 *
 * @return   <code>true</code> if cursor positioning succeeded.
 */
JNIEXPORT jboolean JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_setCursorNative
  (JNIEnv *env, jobject jo, jint x, jint y)
{
   return setCursorNative(x, y);
}

/**
 * Set the cursor visibility on or off. This method only affects the visibility of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 * 
 * @param   on      when <code>true</code> the cursor will become visible, when <code>false</code>
 *                  the cursor will become invisible.
 *
 * @return   The previous state of the cursor (<code>true</code> means visible).
 */
JNIEXPORT jboolean JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_setCursorStatus
  (JNIEnv *env, jobject jo, jboolean on)
{
   return setCursorStatus(env, on);
}

/**
 * Sets the terminal type used by console backend.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   cls     reference to the Java class who called this function (from static context).
 *
 * @param   name    The new terminal type (must not be <code>null</code>).
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_setTermType
  (JNIEnv *env, jclass  cls, jstring name)
{
   setTermType(env, name);
}

/**
 * Save the current state of the console mode and switch to TTY mode temporarily.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_suspend
  (JNIEnv *env, jobject jo)
{
   suspend();
}

/**
 * Synchronizes the state of the console buffers with native terminal window. Meaning the kind of
 * refresh calls depending on the backend used.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 */
JNIEXPORT void JNICALL Java_com_goldencode_p2j_ui_client_chui_driver_console_ConsoleHelper_syncNative
   (JNIEnv *env, jobject jo)
{
   syncNative();
}
