/*
** Module   : terminal_linux.c
** Abstract : Linux implementation for character mode-based console functions.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130723 Created initial version to implement text console terminal JNI.
*/

#include "terminal.h"

// pointer to the structure of the interactive terminal
static SCREEN* terminal = NULL;

static int hascolors = 0;	      // set when we initialize
static int colors_started = 0;	// set when we call start_color().

// The array of the special symbols to draw the horizontal and vertical lines.
#define ACS_MAX 18
static jint acs_result[ACS_MAX];

// The array of the character attributes to use in text drawing
#define ATTRIBUTE_MAX 9
static jint attribute_result[ATTRIBUTE_MAX] =
   {A_NORMAL, A_STANDOUT, A_UNDERLINE, A_REVERSE, A_BLINK, A_DIM, A_BOLD, A_ALTCHARSET, A_INVIS};

// terminal size at the moment of initialization of the library 
extern int base_rows;
extern int base_cols;

/**
 * Throws the exception on the java side for given class name.
 *
 * @param   env        pointer to a structure that provides access to JNI environment.
 *
 * @param   name       the name of the target class for exception.
 * 
 * @param   msg        the message content of the exception.
 */
void throwException(JNIEnv* env, char* name, char* msg)
{
   jclass jcls = (*env)->FindClass(env, name);
   
   if (jcls != NULL)
   {
      (*env)->ThrowNew(env, jcls, msg);
   }
}

/** 
 * This function outputs the screenful of data at once using an array of codes where every
 * character position is either a skip indicator (-1) or a fully prepared code with character,
 * attribute and color included. The final step is the cursor positioning.
 *
 * @param   env      pointer to a structure that provides access to JNI environment.
 *
 * @param   arr      the array of characters and attributes to output. The array is supposed to
 *                   provide an item per every screen position.  This item is either -1 if this
 *                   position is masked or a code which goes directly to ncurses - a combination
 *                   of character, attribute and color.
 *
 * @param   cx       final cursor position, column.
 *
 * @param   cy       final cursor position, row.
 */
JNIEXPORT void JNICALL addArrayNative(JNIEnv *env, jintArray arr, jint cx, jint cy)
{
   int i, j, chr;
   jboolean isCopy = JNI_FALSE;

   jint* arrp = (*env)->GetIntArrayElements(env, arr, &isCopy);
   jint* arp  = arrp;

   if (arp == NULL)
   {
      return;
   }

   for (i = 0; i < base_rows; i++)
   {
      for (j = 0; j < base_cols; j++)
      {
         chr = *arp++;

         if (chr == -1)
         {
            continue;
         }

         move(i, j);
         addch(chr);
       }
   }

   move(cy, cx);
   refresh();

   (*env)->ReleaseIntArrayElements(env, arr, arrp, JNI_ABORT);
}

/**
 * Add a string to the virtual terminal at the current cursor position.
 *
 * @param   env        pointer to a structure that provides access to JNI environment.
 *
 * @param   jstr       The array of characters to output.
 * 
 * @param   len        the lenght of the string to draw in characters.
 *
 * @param   attrib     the attribute to be used while drawing.
 *
 * @param   colorpair  the color value to be used while drawing.
 */
JNIEXPORT void JNICALL addStringNative(JNIEnv *env, jbyteArray jstr, jint len,
                                       jint attrib, jint colorpair)
{
   int i;
   int attr = attrib;

   if (colors_started)
   {
      attr |= COLOR_PAIR(colorpair);
   }
   
   if (len <= 0)
   {
      return;
   }
   
   jbyte* chrs = (jbyte*) malloc(len * sizeof(jbyte));
   
   if (chrs == NULL)
   {
      throwException(env, "java/lang/OutOfMemoryError", "Cannot copy data.");
   }
   
   (*env)->GetByteArrayRegion(env, jstr, 0, len, chrs);
   
   for (i = 0; i < len; i++)
   {
      // remove the automatic sign extension (we want to treat this as
      // an unsigned 8-bit number)
      int extended = 0x000000FF & (int) chrs[i];
      addch(extended | attr);
   } 

   (*env)->DeleteLocalRef(env, jstr);
   free(chrs);
}

/**
 * Beeps the tone on the terminal.
 */
JNIEXPORT void JNICALL beepNative()
{
   beep();
}

/**
 * Clears the current console screen.
 */
JNIEXPORT void JNICALL clearNative()
{
   clear();
}

/**
 * This method is used for getting the line-drawing special characters from underlying native
 * layer.
 *
 * @param   index   The index of the character to return. 
 *
 * @return   The character at the given index from internal array.
 */
JNIEXPORT jint JNICALL getACSchar(jint index)
{
   if (index >= 0 && index < ACS_MAX)
   {
      return acs_result[index];
   }
   else
   {
      return (jint) -1;
   }
}

/**
 * This method is used for getting the attributes from underlying native layer.
 *
 * @param   index   the index of the attribute to return. 
 *
 * @return   The attribute at the given index from internal array.
 */
JNIEXPORT jint JNICALL getAttribute(jint index)
{
   if (index >= 0 && index < ATTRIBUTE_MAX)
   {
      return attribute_result[index];
   }
   else
   {
      return (jint) -1;
   }
}

/**
 * Returns the safe termname without null string on return.
 *
 * @return   The terminal type or "" on any failure.
 */
char* safe_termname()
{
   char* term = termname();
   return (term == NULL) ? "" : term;
}

/**
 * Gets the terminal type reported by console native backend.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @return   The terminal type or an empty string on any failure.
 */
JNIEXPORT jstring JNICALL getTermType(JNIEnv *env)
{
   return (*env)->NewStringUTF(env, safe_termname());
}

/**
 * Get X coordinate of the cursor.
 *
 * @return   The integer value of the X coordinate.
 */
JNIEXPORT jint JNICALL getx()
{
   int x, y;

   getyx(stdscr, y, x);
   return x;
}

/**
 * Get Y coordinate of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The integer value of the Y coordinate.
 */
JNIEXPORT jint JNICALL gety()
{
   int x, y;

   getyx(stdscr, y, x);
   return y;
}

/**
 * Check if the terminal is able to display colors or not.
 *
 * @return   <code>true</code> if display is color capable, <code>false</code> otherwise.
 */
JNIEXPORT jboolean JNICALL hasColors()
{
   return has_colors();
}

/**
 * This method performs console initialization on the native side.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 */
JNIEXPORT void JNICALL initConsole(JNIEnv *env)
{

   char *termname;

   termname = getenv("TERM");
    
   if (termname == NULL || *termname == '\0')
   {
      termname = "vt100";
   }
   
   // use this instead of initscr because there is no way to query the
   // current terminal and when you use multiple terms, newterm overwrites
   // the current terminal... so you can never get it if you don't use
   // newterm directly
   terminal = newterm(termname, stdout, stdin);

   if (terminal == NULL)
   {
      char* spec   = "Interactive terminal of type %s cannot be initialized.";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 1); 
      sprintf(errmsg, spec, termname);
      throwException(env, "java/lang/IllegalStateException", errmsg);
      return;
   }
    
   // ACS_* are not constans and can not be included into static array declarations and should
   // be calculated after new terminal has been created.
   acs_result[0]  = ACS_ULCORNER;
	acs_result[1]  = ACS_LLCORNER;
	acs_result[2]  = ACS_URCORNER;
	acs_result[3]  = ACS_LRCORNER;
	acs_result[4]  = ACS_LTEE;
	acs_result[5]  = ACS_RTEE;
	acs_result[6]  = ACS_BTEE;
	acs_result[7]  = ACS_TTEE;
	acs_result[8]  = ACS_HLINE;
	acs_result[9]  = ACS_VLINE;
	acs_result[10] = ACS_PLUS;
	acs_result[11] = ACS_S1;
	acs_result[12] = ACS_S9;
	acs_result[13] = ACS_DIAMOND;
	acs_result[14] = ACS_CKBOARD;
	acs_result[15] = ACS_DEGREE;
	acs_result[16] = ACS_PLMINUS;
	acs_result[17] = ACS_BULLET;
    
   keypad(stdscr, TRUE);      // enable keyboard mapping
   timeout(500);              // wait up to 500msec for input
   noecho();                  // don't echo input
   raw();
   nonl();
   auto_getch_refresh(FALSE); // disable sync during getch()
   def_prog_mode();           // save our state (after our mode changes)

   hascolors = has_colors();

   getmaxyx(stdscr, base_rows, base_cols);

   atexit((void (*)(void))endwin);
}

/**
 * This method reads the key from keyboard by the special reading thread.
 *
 * @return   The integer code of the key that was read or -1 if there is no character to read.
 */
JNIEXPORT jint JNICALL readKey()
{
   return (jint) getch();
}

/**
 * Reset terminal settings.  May be called only if suspend() was previously called. 
 */
JNIEXPORT void JNICALL resetMode()
{
   reset_prog_mode();
}

/**
 * Re-activates the console mode after ending of the previous suspension state.
 */
JNIEXPORT void JNICALL resume()
{
   reset_prog_mode();          /* Return to the previous tty mode*/
                               /* stored by def_prog_mode()      */
   doupdate();                 /* Do refresh() to restore the    */
                               /* Screen contents                */
}

/**
 * Set absolute cursor position
 *
 * @param   x       new X coordinate of the cursor.
 *
 * @param   y       new Y coordinate of the cursor.
 *
 * @return   <code>true</code> if cursor positioning succeeded.
 */
JNIEXPORT jboolean JNICALL setCursorNative(jint x, jint y)
{
   int localX, localY;
   move(y, x);

   // Set current cursor position
   getyx(stdscr, localY, localX);
    
   return localY == y && localX == x;
}

/**
 * Set the cursor visibility on or off. This method only affects the visibility of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   on      when <code>true</code> the cursor will become visible, when <code>false</code>
 *                  the cursor will become invisible.
 *
 * @return   The previous state of the cursor (<code>true</code> means visible).
 */
JNIEXPORT jboolean JNICALL setCursorStatus(JNIEnv *env, jboolean on)
{
   int rc = ERR;
   jboolean ret;
   
   if( on )
      rc = curs_set( 1 );
   else
      rc = curs_set( 0 );
   
   if ( rc == 1 || rc == 2 )
      ret = true;
   else if ( rc == 0 )
      ret = false;
   else
   {
      char* spec;
      char* termname = getenv("TERM");
      if ( rc == ERR )
         return true;
      else
         spec = "Unexpected return from curs_set() for terminal of type %s (%d).";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 32); 
      sprintf(errmsg, spec, termname, rc);
      throwException(env, "java/lang/IllegalStateException", errmsg);
   }
   
   return ret;
}

/**
 * Simple reset of terminal type.
 *
 * @param   env       pointer to a structure that provides access to JNI environment.
 *
 * @param   termname  The name of the terminal to be restarted.
 */
void restartTerminal(JNIEnv *env, char* termname)
{
   int   err  = 0;
   int   rc   = restartterm(termname, fileno(stdout), &err); 
   
   if (rc == ERR)
   {
      char* spec   = "Failure resetting terminal of type %s (error %i).";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 15); 
      sprintf(errmsg, spec, termname, err);
      throwException(env, "java/lang/IllegalStateException", errmsg);
   }
}

/**
 * Sets the terminal type used by console backend.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   name    The new terminal type (must not be <code>null</code>).
 */
JNIEXPORT void JNICALL setTermType(JNIEnv *env, jstring name)
{
   char* termname = (char*) (*env)->GetStringUTFChars(env, name, NULL);
   
   if (termname == NULL)
   {
      // out of memory!
      exit(-1);
   }
   
   restartTerminal(env, strdup(termname));
   
   (*env)->ReleaseStringUTFChars(env, name, termname);
}

/**
 * Save the current state of the console mode and switch to TTY mode temporarily.
 */
JNIEXPORT void JNICALL suspend()
{
   def_prog_mode();            /* Save the tty modes             */
   endwin();                   /* End curses mode temporarily    */
}

/**
 * Synchronizes the state of the console buffers with native terminal window. Meaning the kind of
 * refresh calls depending on the backend used.
 */
JNIEXPORT void JNICALL syncNative()
{
   refresh();    
}
